@@ -20,6 +20,7 @@
 #include "util/macro.h"
 #include "file/file.h"
 #include "util/bits.h"
+#include "extdata_parse.h"
 #include "clpi_parse.h"
 
 #include <stdlib.h>
@@ -29,8 +30,6 @@
 #define CLPI_SIG2A ('0' << 24 | '2' << 16 | '0' << 8 | '0')
 #define CLPI_SIG2B ('0' << 24 | '1' << 16 | '0' << 8 | '0')
 
-static int clpi_verbose = 0;
-
 static void
 _human_readable_sig(char *sig, uint32_t s1, uint32_t s2)
 {
@@ -64,6 +63,7 @@ _parse_stream_attr(BITSTREAM *bits, CLPI_PROG_STREAM *ss)
         case 0x02:
         case 0xea:
         case 0x1b:
+        case 0x20:
             ss->format = bs_read(bits, 4);
             ss->rate   = bs_read(bits, 4);
             ss->aspect = bs_read(bits, 4);
@@ -221,20 +221,19 @@ _parse_sequence(BITSTREAM *bits, CLPI_CL *cl)
 }
 
 static int
-_parse_program(BITSTREAM *bits, CLPI_CL *cl)
+_parse_program(BITSTREAM *bits, CLPI_PROG_INFO *program)
 {
     int ii, jj;
 
-    bs_seek_byte(bits, cl->program_info_start_addr);
     // Skip the length field, and a reserved byte
     bs_skip(bits, 5 * 8);
     // Then get the number of sequences
-    cl->program.num_prog = bs_read(bits, 8);
+    program->num_prog = bs_read(bits, 8);
 
     CLPI_PROG *progs;
-    progs = malloc(cl->program.num_prog * sizeof(CLPI_PROG));
-    cl->program.progs = progs;
-    for (ii = 0; ii < cl->program.num_prog; ii++) {
+    progs = malloc(program->num_prog * sizeof(CLPI_PROG));
+    program->progs = progs;
+    for (ii = 0; ii < program->num_prog; ii++) {
         progs[ii].spn_program_sequence_start = bs_read(bits, 32);
         progs[ii].program_map_pid            = bs_read(bits, 16);
         progs[ii].num_streams                = bs_read(bits, 8);
@@ -253,6 +252,14 @@ _parse_program(BITSTREAM *bits, CLPI_CL *cl)
     return 1;
 }
 
+static int
+_parse_program_info(BITSTREAM *bits, CLPI_CL *cl)
+{
+    bs_seek_byte(bits, cl->program_info_start_addr);
+
+    return _parse_program(bits, &cl->program);
+}
+
 static int
 _parse_ep_map_stream(BITSTREAM *bits, CLPI_EP_MAP_ENTRY *ee)
 {
@@ -286,29 +293,28 @@ _parse_ep_map_stream(BITSTREAM *bits, CLPI_EP_MAP_ENTRY *ee)
 }
 
 static int
-_parse_cpi(BITSTREAM *bits, CLPI_CL *cl)
+_parse_cpi(BITSTREAM *bits, CLPI_CPI *cpi)
 {
     int ii;
     uint32_t ep_map_pos, len;
 
-    bs_seek_byte(bits, cl->cpi_start_addr);
     len = bs_read(bits, 32);
     if (len == 0) {
         return 1;
     }
 
     bs_skip(bits, 12);
-    cl->cpi.type = bs_read(bits, 4);
+    cpi->type = bs_read(bits, 4);
     ep_map_pos = bs_pos(bits) >> 3;
 
     // EP Map starts here
     bs_skip(bits, 8);
-    cl->cpi.num_stream_pid = bs_read(bits, 8);
+    cpi->num_stream_pid = bs_read(bits, 8);
 
     CLPI_EP_MAP_ENTRY *entry;
-    entry = malloc(cl->cpi.num_stream_pid * sizeof(CLPI_EP_MAP_ENTRY));
-    cl->cpi.entry = entry;
-    for (ii = 0; ii < cl->cpi.num_stream_pid; ii++) {
+    entry = malloc(cpi->num_stream_pid * sizeof(CLPI_EP_MAP_ENTRY));
+    cpi->entry = entry;
+    for (ii = 0; ii < cpi->num_stream_pid; ii++) {
         entry[ii].pid                      = bs_read(bits, 16);
         bs_skip(bits, 10);
         entry[ii].ep_stream_type           = bs_read(bits, 4);
@@ -316,12 +322,20 @@ _parse_cpi(BITSTREAM *bits, CLPI_CL *cl)
         entry[ii].num_ep_fine              = bs_read(bits, 18);
         entry[ii].ep_map_stream_start_addr = bs_read(bits, 32) + ep_map_pos;
     }
-    for (ii = 0; ii < cl->cpi.num_stream_pid; ii++) {
-        _parse_ep_map_stream(bits, &cl->cpi.entry[ii]);
+    for (ii = 0; ii < cpi->num_stream_pid; ii++) {
+        _parse_ep_map_stream(bits, &cpi->entry[ii]);
     }
     return 1;
 }
 
+static int
+_parse_cpi_info(BITSTREAM *bits, CLPI_CL *cl)
+{
+    bs_seek_byte(bits, cl->cpi_start_addr);
+
+    return _parse_cpi(bits, &cl->cpi);
+}
+
 static uint32_t
 _find_stc_spn(const CLPI_CL *cl, uint8_t stc_id)
 {
@@ -532,6 +546,74 @@ clpi_access_point(const CLPI_CL *cl, uint32_t pkt, int next, int angle_change, u
     return coarse_spn + entry->fine[jj].spn_ep;
 }
 
+static int
+_parse_extent_start_points(BITSTREAM *bits, CLPI_EXTENT_START *es)
+{
+    unsigned int ii;
+
+    bs_skip(bits, 32); // length
+    es->num_point = bs_read(bits, 32);
+
+    es->point = malloc(es->num_point * sizeof(uint32_t));
+
+    for (ii = 0; ii < es->num_point; ii++) {
+        es->point[ii] = bs_read(bits, 32);
+    }
+
+    return 1;
+}
+
+static int _parse_clpi_extension(BITSTREAM *bits, int id1, int id2, void *handle)
+{
+    CLPI_CL *cl = (CLPI_CL*)handle;
+
+    if (id1 == 2) {
+        if (id2 == 4) {
+            // Extent start point
+            return _parse_extent_start_points(bits, &cl->extent_start);
+        }
+        if (id2 == 5) {
+            // ProgramInfo SS
+            return _parse_program(bits, &cl->program_ss);
+        }
+        if (id2 == 6) {
+            // CPI SS
+            return _parse_cpi(bits, &cl->cpi_ss);
+        }
+    }
+
+    return 0;
+}
+
+static void
+_clean_program(CLPI_PROG_INFO *p)
+{
+    int ii;
+
+    for (ii = 0; ii < p->num_prog; ii++) {
+        if (p->progs[ii].streams != NULL) {
+            X_FREE(p->progs[ii].streams);
+        }
+    }
+    X_FREE(p->progs);
+}
+
+static void
+_clean_cpi(CLPI_CPI *cpi)
+{
+    int ii;
+
+    for (ii = 0; ii < cpi->num_stream_pid; ii++) {
+        if (cpi->entry[ii].coarse != NULL) {
+            X_FREE(cpi->entry[ii].coarse);
+        }
+        if (cpi->entry[ii].fine != NULL) {
+            X_FREE(cpi->entry[ii].fine);
+        }
+    }
+    X_FREE(cpi->entry);
+}
+
 void
 clpi_free(CLPI_CL *cl)
 {
@@ -552,38 +634,24 @@ clpi_free(CLPI_CL *cl)
         X_FREE(cl->sequence.atc_seq);
     }
 
-    for (ii = 0; ii < cl->program.num_prog; ii++) {
-        if (cl->program.progs[ii].streams != NULL) {
-            X_FREE(cl->program.progs[ii].streams);
-        }
-    }
-    if (cl->program.progs != NULL) {
-        X_FREE(cl->program.progs);
-    }
+    _clean_program(&cl->program);
+    _clean_cpi(&cl->cpi);
+
+    X_FREE(cl->extent_start.point);
+
+    _clean_program(&cl->program_ss);
+    _clean_cpi(&cl->cpi_ss);
 
-    for (ii = 0; ii < cl->cpi.num_stream_pid; ii++) {
-        if (cl->cpi.entry[ii].coarse != NULL) {
-            X_FREE(cl->cpi.entry[ii].coarse);
-        }
-        if (cl->cpi.entry[ii].fine != NULL) {
-            X_FREE(cl->cpi.entry[ii].fine);
-        }
-    }
-    if (cl->cpi.entry != NULL) {
-        X_FREE(cl->cpi.entry);
-    }
     X_FREE(cl);
 }
 
 static CLPI_CL*
-_clpi_parse(const char *path, int verbose)
+_clpi_parse(const char *path)
 {
     BITSTREAM  bits;
     BD_FILE_H *fp;
     CLPI_CL   *cl;
 
-    clpi_verbose = verbose;
-
     cl = calloc(1, sizeof(CLPI_CL));
     if (cl == NULL) {
         return NULL;
@@ -602,6 +670,14 @@ _clpi_parse(const char *path, int verbose)
         clpi_free(cl);
         return NULL;
     }
+
+    if (cl->ext_data_start_addr > 0) {
+        bdmv_parse_extension_data(&bits,
+                                   cl->ext_data_start_addr,
+                                   _parse_clpi_extension,
+                                   cl);
+    }
+
     if (!_parse_clipinfo(&bits, cl)) {
         file_close(fp);
         clpi_free(cl);
@@ -612,12 +688,12 @@ _clpi_parse(const char *path, int verbose)
         clpi_free(cl);
         return NULL;
     }
-    if (!_parse_program(&bits, cl)) {
+    if (!_parse_program_info(&bits, cl)) {
         file_close(fp);
         clpi_free(cl);
         return NULL;
     }
-    if (!_parse_cpi(&bits, cl)) {
+    if (!_parse_cpi_info(&bits, cl)) {
         file_close(fp);
         clpi_free(cl);
         return NULL;
@@ -627,20 +703,20 @@ _clpi_parse(const char *path, int verbose)
 }
 
 CLPI_CL*
-clpi_parse(const char *path, int verbose)
+clpi_parse(const char *path)
 {
-    CLPI_CL *cl = _clpi_parse(path, verbose);
+    CLPI_CL *cl = _clpi_parse(path);
 
     /* if failed, try backup file */
     if (!cl) {
-        int   len    = strlen(path);
+        size_t len   = strlen(path);
         char *backup = malloc(len + 8);
 
         strncpy(backup, path, len - 18);
         strcpy(backup + len - 18, "BACKUP/");
         strcpy(backup + len - 18 + 7, path + len - 18);
 
-        cl = _clpi_parse(backup, verbose);
+        cl = _clpi_parse(backup);
 
         X_FREE(backup);
     }
