@@ -1,6 +1,6 @@
 /*
  * This file is part of libbluray
- * Copyright (C) 2010  hpi1
+ * Copyright (C) 2010-2012  Petri Hintukainen <phintuka@users.sourceforge.net>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -220,7 +220,7 @@ static int _get_event(HDMV_VM *p, HDMV_EVENT *ev)
     return -1;
 }
 
-static int _queue_event(HDMV_VM *p, uint32_t event, uint32_t param)
+static int _queue_event(HDMV_VM *p, hdmv_event_e event, uint32_t param)
 {
     unsigned i;
     for (i = 0; i < sizeof(p->event) / sizeof(p->event[0]) - 1; i++) {
@@ -405,7 +405,9 @@ static int _resume_object(HDMV_VM *p, int psr_restore)
 
     p->suspended_object = NULL;
 
-    BD_DEBUG(DBG_HDMV, "resuming object %p at %d\n", p->object, p->pc);
+    BD_DEBUG(DBG_HDMV, "resuming object %ld at %d\n",
+             (long)((p->movie_objects->objects - p->object) / sizeof(p->movie_objects->objects[0])),
+             p->pc);
 
     _queue_event(p, HDMV_EVENT_PLAY_STOP, 0);
 
@@ -548,7 +550,13 @@ static int _play_stop(HDMV_VM *p)
     }
 
     BD_DEBUG(DBG_HDMV, "_play_stop()\n");
-    _queue_event(p, HDMV_EVENT_PLAY_STOP, 0);
+    _queue_event(p, HDMV_EVENT_PLAY_STOP, 1);
+
+    /* terminate IG object. Continue executing movie object.  */
+    if (_resume_from_play_pl(p) < 0) {
+        BD_DEBUG(DBG_HDMV|DBG_CRIT, "_play_stop(): resuming movie object failed !\n");
+        return -1;
+    }
 
     return 0;
 }
@@ -651,6 +659,18 @@ static void _set_sec_stream(HDMV_VM *p, uint32_t dst, uint32_t src)
     bd_psr_unlock(p->regs);
 }
 
+static void _set_stream_ss(HDMV_VM *p, uint32_t dst, uint32_t src)
+{
+    BD_DEBUG(DBG_HDMV, "_set_stream_ss(0x%x, 0x%x)\n", dst, src);
+
+    if (!(bd_psr_read(p->regs, PSR_3D_STATUS) & 1)) {
+        BD_DEBUG(DBG_HDMV, "_set_stream_ss ignored (PSR22 indicates 2D mode)\n");
+        return;
+    }
+
+    BD_DEBUG(DBG_HDMV, "_set_stream_ss(0x%x, 0x%x) unimplemented\n", dst, src);
+}
+
 /*
  * SET/SYSTEM navigation control
  */
@@ -711,6 +731,33 @@ static void _popup_off(HDMV_VM *p)
     }
 }
 
+/*
+ * SET/SYSTEM 3D mode
+ */
+
+static void _set_output_mode(HDMV_VM *p, uint32_t dst)
+{
+    if ((bd_psr_read(p->regs, PSR_PROFILE_VERSION) & 0x130240) != 0x130240) {
+        BD_DEBUG(DBG_HDMV, "_set_output_mode ignored (not running as profile 5 player)\n");
+        return;
+    }
+
+    bd_psr_lock(p->regs);
+
+    uint32_t psr22 = bd_psr_read(p->regs, PSR_3D_STATUS);
+
+    /* update output mode (bit 0). PSR22 bits 1 and 2 are subtitle alignment (_set_stream_ss()) */
+    if (dst & 1) {
+      psr22 |= 1;
+    } else {
+      psr22 &= ~1;
+    }
+
+    bd_psr_write(p->regs, PSR_3D_STATUS, psr22);
+
+    bd_psr_unlock(p->regs);
+}
+
 /*
  * navigation timer
  */
@@ -739,6 +786,8 @@ static void _set_nv_timer(HDMV_VM *p, uint32_t dst, uint32_t src)
       return;
   }
 
+  BD_DEBUG(DBG_HDMV | DBG_CRIT, "_set_nv_timer(): navigation timer not implemented !\n");
+
   /* set expiration time */
   p->nv_timer.time = time(NULL);
   p->nv_timer.time += timeout;
@@ -830,14 +879,14 @@ static inline uint32_t ADD_u32(uint32_t a, uint32_t b)
 {
   /* overflow -> saturate */
   uint64_t result = (uint64_t)a + b;
-  return result < 0xffffffff ? result : 0xffffffff;
+  return result < 0xffffffff ? (uint32_t)result : 0xffffffff;
 }
 
 static inline uint32_t MUL_u32(uint32_t a, uint32_t b)
 {
   /* overflow -> saturate */
   uint64_t result = (uint64_t)a * b;
-  return result < 0xffffffff ? result : 0xffffffff;
+  return result < 0xffffffff ? (uint32_t)result : 0xffffffff;
 }
 
 /*
@@ -978,6 +1027,8 @@ static int _hdmv_step(HDMV_VM *p)
                         case INSN_POPUP_OFF:       _popup_off      (p);           break;
                         case INSN_STILL_ON:        _set_still_mode (p,   1);      break;
                         case INSN_STILL_OFF:       _set_still_mode (p,   0);      break;
+                        case INSN_SET_OUTPUT_MODE: _set_output_mode(p, dst);      break;
+                        case INSN_SET_STREAM_SS:   _set_stream_ss  (p, dst, src); break;
                         default:
                             BD_DEBUG(DBG_HDMV|DBG_CRIT, "[unknown SETSYSTEM option in opcode 0x%08x] ", *(uint32_t*)insn);
                             break;
